"""Run checks."""

import datetime as dt
import logging
import signal
import time
import types
from functools import partial
from threading import Event

import click
from regal.exceptions import NssetDoesNotExist
from sqlalchemy import delete, update

import dnscheck
from dnscheck.db import DomainCheck, NssetCheck, atomic
from dnscheck.process import domain_checks, nsset_checks, nssets
from dnscheck.settings import APP
from dnscheck.utils import StopProcess, async_to_sync, defer_errors, run_unless_stopped

_LOGGER = logging.getLogger(__name__)

_DATETIME_FORMATS = ("%Y-%m-%dT%H:%M:%S%z", "%Y-%m-%d %H:%M:%S%z")


def _signal_handler(signum: int, frame: types.FrameType | None, *, stop_event: Event) -> None:  # pragma: no cover
    """Capture signal and set stop event."""
    _LOGGER.info("Captured signal %s, stopping dnscheck daemon", signal.Signals(signum).name)
    stop_event.set()


@click.group()
@click.version_option(version=dnscheck.__version__, message="fred-dnscheck %(version)s")
@click.option("--config", type=click.Path(exists=True, dir_okay=False), help="Config file.")
def cli(config: str) -> None:
    """DNScheck CLI."""
    APP.load(config)


@cli.command()
@click.option(
    "--batch-size", default=1000, show_default=True, help="How many nssets should be processed simultaneously."
)
def sync(batch_size: int) -> None:
    """Sync nssets with domain registry."""
    nssets.sync(batch_size=batch_size)


@cli.command()
@click.option("--loop-delay", default=5, show_default=True, help="Delay in seconds between processing loop runs.")
def run(loop_delay: int) -> None:
    """Run main process loop."""
    _LOGGER.info("Starting dnscheck daemon")
    stop_event = Event()
    signal.signal(signal.SIGTERM, partial(_signal_handler, stop_event=stop_event))

    _run = run_unless_stopped(defer_errors, stop_event=stop_event)
    try:
        while True:
            _run(domain_checks.finish)()
            _run(nsset_checks.start)()
            _run(domain_checks.start)()
            _run(nsset_checks.finish)()
            _run(time.sleep)(loop_delay)
    except KeyboardInterrupt:
        _LOGGER.info("Stopping dnscheck daemon")
    except StopProcess:
        _LOGGER.info("Dnscheck daemon has been gracefully stopped")


@cli.command()
@click.option(
    "--before",
    type=click.DateTime(formats=_DATETIME_FORMATS),
    required=True,
    help="Domain check results older than this timezone-aware datetime will be cleaned.",
)
def clean_results(before: dt.datetime) -> None:
    """Clean old domain check results.

    The results are replaced with NULL.
    The domain checks themselves are preserved.
    """
    with atomic() as db:
        db.execute(update(DomainCheck).where(DomainCheck.finished_at < before).values(results=None))


@cli.command()
@click.option(
    "--before",
    type=click.DateTime(formats=_DATETIME_FORMATS),
    required=True,
    help="Checks older than this timezone-aware datetime will be cleaned.",
)
def clean_checks(before: dt.datetime) -> None:
    """Clean old nsset checks and domain checks."""
    with atomic() as db:
        result = db.execute(delete(NssetCheck).where(NssetCheck.finished_at < before))

    click.echo(f"Deleted nsset checks: {result.rowcount}")


@cli.command()
@click.argument("handle", nargs=-1)
def check(handle: tuple[str]) -> None:
    """Start NSSET checks."""
    for nsset_handle in handle:
        try:
            nsset_id = async_to_sync(APP.nsset_client.get_nsset_id)(nsset_handle)
        except NssetDoesNotExist:
            click.echo(f"ERROR: Nsset {nsset_handle} does not exist")
            continue
        nsset_checks.start_one(nsset_id)
        click.echo(f"INFO: Started nsset check of {nsset_handle}")
