"""Database utilities."""

import sqlite3
from collections.abc import Generator
from contextlib import contextmanager
from functools import lru_cache
from typing import Any, cast

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase, Session, sessionmaker

from dnscheck.settings import APP


@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection: Any, connection_record: Any) -> None:
    """Enable SQLite foreign keys.

    See https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support
    """
    if isinstance(dbapi_connection, sqlite3.Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()


@lru_cache
def get_engine() -> Engine:
    """Return engine as defined by settings."""
    engine = create_engine(APP.settings.db_connection, echo=APP.settings.db_echo, poolclass=APP.settings.db_poolclass)
    # Redirect all queries to schema defined in settings.
    # See https://docs.sqlalchemy.org/en/20/core/connections.html#translation-of-schema-names
    engine.update_execution_options(schema_translate_map={None: APP.settings.db_schema})
    return engine


@lru_cache
def get_session_cls(expire_on_commit: bool = True) -> type[Session]:
    """Return session class for defined engine."""
    return cast(type[Session], sessionmaker(bind=get_engine(), expire_on_commit=expire_on_commit))


@contextmanager
def atomic(expire_on_commit: bool = True) -> Generator[Session, None, None]:
    """Context manager which wraps content into a transaction."""
    session = get_session_cls(expire_on_commit=expire_on_commit)()
    try:
        yield session
        session.commit()
    except BaseException:
        session.rollback()
        raise
    finally:
        session.close()


class Base(DeclarativeBase):
    """Base for database models."""
