"""DNScheck database models."""

import datetime as dt

from sqlalchemy import JSON, Enum, ForeignKey, Index, func
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy_utc import UtcDateTime, utcnow

from dnscheck.constants import DomainCheckStatus, NssetCheckProfile, NssetCheckStatus, TestResultLevel

from .common import Base


class Nsset(Base):
    """Nsset table."""

    __tablename__ = "dnscheck_nsset"

    id: Mapped[int] = mapped_column(primary_key=True)
    nsset_id: Mapped[str] = mapped_column(unique=True, index=True)
    handle: Mapped[str]
    last_synced: Mapped[dt.datetime] = mapped_column(UtcDateTime(), server_default=utcnow(), onupdate=utcnow())


class NssetCheck(Base):
    """Nsset check.

    Attributes:
        handle: Nsset handle.
        nameservers: List of nsset nameservers.
        profile: Check profile based on nsset ``resultlevel``.
    """

    __tablename__ = "dnscheck_nsset_check"

    id: Mapped[int] = mapped_column(primary_key=True)
    nsset_id: Mapped[str] = mapped_column(index=True)
    nsset_history_id: Mapped[str] = mapped_column(index=True)
    handle: Mapped[str]
    nameservers: Mapped[list[dict]] = mapped_column(JSON)
    profile: Mapped[NssetCheckProfile] = mapped_column(Enum(NssetCheckProfile), default=NssetCheckProfile.DEFAULT)

    status: Mapped[NssetCheckStatus] = mapped_column(Enum(NssetCheckStatus), default=NssetCheckStatus.STARTED)
    started_at: Mapped[dt.datetime] = mapped_column(UtcDateTime(), server_default=func.now(), index=True)
    finished_at: Mapped[dt.datetime | None] = mapped_column(UtcDateTime())

    domain_checks: Mapped[list["DomainCheck"]] = relationship(
        "DomainCheck", back_populates="nsset_check", cascade="all, delete-orphan"
    )

    __table_args__ = (
        Index(
            "dnscheck_nsset_check_nsset_id_active",
            "nsset_id",
            unique=True,
            postgresql_where=(status != NssetCheckStatus.FINISHED),
        ).ddl_if(dialect="postgresql"),
        Index(
            "dnscheck_nsset_check_nsset_id_active",
            "nsset_id",
            unique=True,
            sqlite_where=(status != NssetCheckStatus.FINISHED),
        ).ddl_if(dialect="sqlite"),
    )


class DomainCheck(Base):
    """Domain check.

    Attributes:
        domain: FQDN of checked domain.
        test_id: Zonemaster test id.
        results: Full Zonemaster test results.
        summary: Summary of test results severity.
    """

    __tablename__ = "dnscheck_domain_check"

    id: Mapped[int] = mapped_column(primary_key=True)
    nsset_check_id: Mapped[int] = mapped_column(ForeignKey(NssetCheck.id, ondelete="CASCADE"))
    domain_id: Mapped[str | None]  # TODO: Remove None when possible
    domain: Mapped[str]
    ds_info: Mapped[list[dict] | None] = mapped_column(JSON)
    test_id: Mapped[str | None]

    status: Mapped[DomainCheckStatus] = mapped_column(Enum(DomainCheckStatus), default=DomainCheckStatus.CREATED)
    started_at: Mapped[dt.datetime | None] = mapped_column(UtcDateTime())
    finished_at: Mapped[dt.datetime | None] = mapped_column(UtcDateTime())
    results: Mapped[list[dict] | None] = mapped_column(JSON)
    summary: Mapped[dict[TestResultLevel, int] | None] = mapped_column(JSON)

    nsset_check: Mapped[NssetCheck] = relationship(NssetCheck, back_populates="domain_checks")
