"""Process domain checks."""

import datetime as dt
import logging
from collections import Counter
from typing import cast

from fred_types import DomainId
from sqlalchemy import func, select
from sqlalchemy.orm import selectinload

from dnscheck.constants import DomainCheckStatus
from dnscheck.db import DomainCheck, NssetCheck, atomic
from dnscheck.schema import DSInfo, Nameserver
from dnscheck.settings import APP
from dnscheck.utils import async_to_sync
from dnscheck.zonemaster import ZonemasterError

_LOGGER = logging.getLogger(__name__)


@async_to_sync
async def start() -> None:
    """Start multiple domain checks up to limit."""
    with atomic() as sess:
        stmt_count = select(func.count(DomainCheck.id)).where(DomainCheck.status.in_({DomainCheckStatus.STARTED}))
        domain_checks_running = sess.execute(stmt_count).scalar_one()

        if domain_checks_running < APP.settings.max_domain_checks:
            stmt = (
                select(DomainCheck)
                .join(NssetCheck, DomainCheck.nsset_check_id == NssetCheck.id)
                .where(DomainCheck.status == DomainCheckStatus.CREATED)
                .order_by(NssetCheck.started_at.asc())
                .with_for_update(skip_locked=True)
                .limit(APP.settings.max_domain_checks - domain_checks_running)
                .options(selectinload(DomainCheck.nsset_check))
            )
            domain_checks = sess.execute(stmt).scalars().all()

            for check in domain_checks:
                await start_one(check)


async def start_one(domain_check: DomainCheck) -> None:
    """Start domain check in Zonemaster.

    Args:
        domain_check: DomainCheck instance with status CREATED.
    """
    nameservers = [
        Nameserver(
            fqdn=ns["fqdn"],
            ip_addresses=ns.get("ip_addresses", []),
        )
        for ns in domain_check.nsset_check.nameservers
    ]
    if not domain_check.domain_id:
        # TODO: Remove when DomainCheck.domain_id is no longer nullable
        domain_check.domain_id = await APP.domain_client.get_domain_id(domain_check.domain)
    domain = await APP.domain_client.get_domain_info(DomainId(domain_check.domain_id))
    if domain.keyset:
        keyset = await APP.keyset_client.get_keyset_info(domain.keyset.id)
        ds_info = [DSInfo.from_ds_record(dnskey.ds_record(domain_check.domain)) for dnskey in keyset.dns_keys]
    else:
        ds_info = []

    test_id = APP.zonemaster.start_domain_test(
        fqdn=domain_check.domain,
        nameservers=nameservers,
        ds_info=ds_info,
        profile=domain_check.nsset_check.profile,
    )
    domain_check.ds_info = [ds.model_dump(mode="json") for ds in ds_info]
    domain_check.status = cast(DomainCheckStatus, DomainCheckStatus.STARTED)
    domain_check.test_id = test_id
    domain_check.started_at = dt.datetime.now(tz=dt.UTC)
    _LOGGER.info("Started domain check of %s", domain_check.domain)


def finish() -> None:
    """Finish completed domain checks.

    Find domain checks with state STARTED.
    For each domain check call zonemaster API to check its progress.
    If the test is already finished, fetch the results and finish the domain check.
    """
    with atomic() as db:
        stmt = (
            select(DomainCheck)
            .where(DomainCheck.status == DomainCheckStatus.STARTED)
            .with_for_update(skip_locked=True)
            .options(selectinload(DomainCheck.nsset_check))
        )
        domain_checks = db.execute(stmt).scalars()
        for domain_check in domain_checks:
            try:
                progress = APP.zonemaster.test_progress(cast(str, domain_check.test_id))
                if progress >= 100:
                    finish_one(domain_check)
            except ZonemasterError:
                # Reset the domain check
                _LOGGER.warning("Resetting domain check of %s due to Zonemaster error", domain_check.domain)
                domain_check.status = DomainCheckStatus.CREATED
                domain_check.started_at = None
                domain_check.test_id = None


def finish_one(domain_check: DomainCheck) -> None:
    """Finish one completed domain check.

    Args:
        domain_check: DomainCheck instance with status STARTED and test progress 100%.
    """
    test_results = APP.zonemaster.get_test_results(cast(str, domain_check.test_id))
    domain_check.status = DomainCheckStatus.FINISHED
    domain_check.finished_at = dt.datetime.now(tz=dt.UTC)
    domain_check.results = test_results
    domain_check.summary = dict(Counter(res["level"] for res in test_results))
    _LOGGER.info("Finished domain check of %s", domain_check.domain)
