"""DNScheck settings."""

import datetime as dt
from functools import cached_property
from typing import Annotated

from frgal import make_credentials
from hermes import EmailMessengerClient
from pydantic import BeforeValidator, HttpUrl, ImportString, PositiveInt, TypeAdapter
from regal import ContactClient, DomainClient, KeysetClient, NssetClient
from setapp import BaseApplication, BaseSettings, SettingsConfigDict

from .zonemaster import ZonemasterClient

# NOTE: this is a workaround that validates the URL, but returns str
# See https://github.com/pydantic/pydantic/discussions/6395
_http_url_adapter = TypeAdapter(HttpUrl)
HttpUrlString = Annotated[str, BeforeValidator(lambda value: str(_http_url_adapter.validate_python(value)))]


class Settings(BaseSettings):
    """App settings.

    Attributes:
        db_connection: Database connection string.
        db_echo: If True, log all SQL queries.
        db_poolclass: Dotted path to database poolclass.
        db_schema: Database schema to use.
        messenger_netloc: Netloc (i.e. host and port) of messenger service.
        messenger_ssl_cert: Path to file with SSL root certificate.
        registry_netloc: Netloc (i.e. host and port) of registry service.
        registry_ssl_cert: Path to file with SSL root certificate.
        zonemaster_api_url: URL of Zonemaster RPC API.

        max_domain_checks: Maximum number of concurrent domain checks.
        min_checks_interval: Minimum delay between checks of the same nsset.
        max_notified_domains: Maximum number of bad domains reported in notification email.
    """

    model_config = SettingsConfigDict(
        config_environ="DNSCHECK_CONFIG",
        config_files=("~/.fred/dnscheck.conf", "/etc/fred/dnscheck.conf"),
        env_prefix="dnscheck_",
    )

    # Service connections
    db_connection: str = "sqlite://"
    db_echo: bool = False
    db_poolclass: ImportString | None = None
    db_schema: str | None = None
    messenger_netloc: str | None = None
    messenger_ssl_cert: str | None = None
    registry_netloc: str
    registry_ssl_cert: str | None = None
    zonemaster_api_url: HttpUrlString = "http://localhost/"

    # Limits
    max_domain_checks: PositiveInt = 30
    max_notified_domains: PositiveInt = 10
    min_check_interval: dt.timedelta = dt.timedelta(days=182)  # 6 months


class Application(BaseApplication[Settings]):
    """Represent application settings and basic setup."""

    @cached_property
    def zonemaster(self) -> ZonemasterClient:
        """Return a Zonemaster client."""
        return ZonemasterClient(self.settings.zonemaster_api_url)

    @cached_property
    def contact_client(self) -> ContactClient:
        """Return a contact client."""
        return ContactClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def domain_client(self) -> DomainClient:
        """Return a domain client."""
        return DomainClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def keyset_client(self) -> KeysetClient:
        """Return a keyset client."""
        return KeysetClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def nsset_client(self) -> NssetClient:
        """Return a nsset client."""
        return NssetClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def email_client(self) -> EmailMessengerClient | None:
        """Return messenger email client."""
        if self.settings.messenger_netloc:
            return EmailMessengerClient(
                self.settings.messenger_netloc,
                make_credentials(self.settings.messenger_ssl_cert),
            )
        else:
            return None

    def clear_cache(self) -> None:
        """Clear all caches."""
        self._del("zonemaster")
        self._del("contact_client")
        self._del("domain_client")
        self._del("keyset_client")
        self._del("nsset_client")
        self._del("email_client")


APP = Application(Settings)
