import os
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path

import pytest
from sqlalchemy import Engine
from sqlalchemy_utils import create_database, database_exists, drop_database

from dnscheck.db.common import Base, get_engine, get_session_cls
from dnscheck.settings import APP

from .utils import override_settings, slugify


@pytest.fixture(autouse=True)
def _load_app():
    if not APP.is_loaded:
        APP.load()


@pytest.fixture(params=["sqlite", "postgres"])
def db_connection(request: pytest.FixtureRequest, worker_id: str, monkeypatch) -> Generator[str, None, None]:
    if request.param == "sqlite":
        yield "sqlite://"

    if request.param == "postgres":
        with monkeypatch.context() as ctx:
            tox_env = slugify(os.getenv("TOX_ENV_NAME", ""))
            ctx.setenv("PGDATABASE", f"dnscheck_test_{tox_env}_{worker_id}")
            yield "postgresql+psycopg://"


@contextmanager
def _db_engine(db_connection: str) -> Generator[Engine, None, None]:
    with override_settings(db_connection=db_connection):
        get_session_cls.cache_clear()
        get_engine.cache_clear()

        engine = get_engine()

        # XXX: This hack is needed until SQLAlchemy-Utils learns how to use PGDATABASE directly.
        # See https://github.com/kvesteri/sqlalchemy-utils/pull/463
        url = engine.url
        if url.get_dialect().name == "postgresql":
            url = url.set(database=os.environ.get("PGDATABASE"))

        if database_exists(url):
            drop_database(url)
        create_database(url)

        yield engine

        drop_database(url)
        get_engine.cache_clear()
        get_session_cls.cache_clear()


@pytest.fixture
def db(db_connection: str) -> Generator[Engine, None, None]:
    with _db_engine(db_connection) as engine:
        with engine.connect() as conn:
            Base.metadata.create_all(bind=conn)
            conn.commit()

        yield engine


# Required by pytest-alembic
@pytest.fixture
def alembic_config():
    return {"file": Path(__file__).parent.parent / "alembic.ini"}


@pytest.fixture
def alembic_engine(db_connection: str) -> Generator[Engine, None, None]:
    with _db_engine(db_connection) as engine:
        yield engine
