import datetime as dt
from unittest.mock import call

import pytest
from fred_types import KeysetRef
from pytest_mock import MockFixture
from sqlalchemy import select
from testfixtures import Comparison as C
from time_machine import TimeMachineFixture

from dnscheck.constants import DomainCheckStatus, NssetCheckProfile, TestResultLevel
from dnscheck.db import DomainCheck, atomic
from dnscheck.process import domain_checks
from dnscheck.schema import DSInfo, Nameserver
from dnscheck.settings import APP
from dnscheck.tests.utils import ModelFactory, RegistryFactory, override_settings
from dnscheck.utils import async_to_sync
from dnscheck.zonemaster import ZonemasterError


def _prepare_start_db(started: int = 0, created: int = 0) -> list[DomainCheck]:
    checks = []
    mf = ModelFactory()
    with atomic(expire_on_commit=False) as sess:
        sess.add_all([mf.nsset(1), mf.nsset_check(1)])
        for i in range(started):
            check = mf.domain_check(f"{i+1}.example.org", status=DomainCheckStatus.STARTED)
            checks.append(check)
            sess.add(check)
        for i in range(started, started + created):
            check = mf.domain_check(f"{i+1}.example.org", status=DomainCheckStatus.CREATED)
            checks.append(check)
            sess.add(check)
    return checks


def test_start(db, mocker: MockFixture):
    mock_start_one = mocker.patch("dnscheck.process.domain_checks.start_one", side_effect=[True, False, True, True])
    _prepare_start_db(0, 4)

    with override_settings(max_domain_checks=10):
        domain_checks.start()

    with atomic() as sess:
        sess.add_all([c.args[0] for c in mock_start_one.mock_calls])
        assert {c.args[0].domain for c in mock_start_one.mock_calls} == {
            "1.example.org",
            "2.example.org",
            "3.example.org",
            "4.example.org",
        }


def test_start_some_running(db, mocker: MockFixture):
    mock_start_one = mocker.patch("dnscheck.process.domain_checks.start_one", return_value=True)
    _prepare_start_db(1, 3)

    with override_settings(max_domain_checks=3):
        domain_checks.start()

    with atomic() as sess:
        sess.add_all([c.args[0] for c in mock_start_one.mock_calls])
        assert (
            len(
                {c.args[0].domain for c in mock_start_one.mock_calls}
                & {"2.example.org", "3.example.org", "4.example.org"}
            )
            == 2
        )


def test_start_no_checks(db, mocker: MockFixture):
    mock_start_one = mocker.patch("dnscheck.process.domain_checks.start_one")
    _prepare_start_db(2, 0)

    with override_settings(max_domain_checks=10):
        domain_checks.start()

    assert mock_start_one.mock_calls == []


def test_start_too_many_checks(db, mocker: MockFixture):
    mock_start_one = mocker.patch("dnscheck.process.domain_checks.start_one")
    _prepare_start_db(2, 2)

    with override_settings(max_domain_checks=2):
        domain_checks.start()

    assert mock_start_one.mock_calls == []


def test_start_one(db, mocker: MockFixture, time_machine: TimeMachineFixture):
    zm_start_test_mock = mocker.patch.object(APP.zonemaster, "start_domain_test", return_value="Gazpacho!")
    rf = RegistryFactory()
    keyset = rf.keyset(1, dns_keys=[rf.dnskey(key=b"Gazpacho!"), rf.dnskey(key=b"Starbug!")])
    domain = rf.domain(1, fqdn="example.org", keyset=KeysetRef(id=keyset.keyset_id, handle=keyset.keyset_handle))
    domain_mock = mocker.patch.object(APP.domain_client, "get_domain_info", return_value=domain)
    keyset_mock = mocker.patch.object(APP.keyset_client, "get_keyset_info", return_value=keyset)

    now = dt.datetime(2000, 1, 1, tzinfo=dt.UTC)
    time_machine.move_to(now)
    dt.datetime.now()  # start the time machine

    nameservers: list[dict] = [
        {"fqdn": "ns1.example.org", "ip_addresses": ["127.0.0.1", "::1"]},
        {"fqdn": "ns2.example.org"},
    ]
    mf = ModelFactory()
    with atomic(expire_on_commit=False) as sess:
        sess.add_all(
            [
                mf.nsset_check(1, nameservers=nameservers),
                dc := mf.domain_check("example.org", domain_id=domain.domain_id),
            ]
        )

    with atomic() as sess:
        sess.add(dc)
        async_to_sync(domain_checks.start_one)(dc)

    with atomic() as sess:
        sess.add(dc)
        sess.refresh(dc)
        assert dc.status == DomainCheckStatus.STARTED
        assert dc.test_id == "Gazpacho!"
        assert dc.started_at.timestamp() == pytest.approx(now.timestamp(), abs=1)

    assert domain_mock.mock_calls == [call(domain.domain_id)]
    assert keyset_mock.mock_calls == [call(keyset.keyset_id)]
    assert zm_start_test_mock.mock_calls == [
        call(
            fqdn="example.org",
            nameservers=[Nameserver(**ns) for ns in nameservers],
            ds_info=[DSInfo.from_ds_record(dnskey.ds_record("example.org")) for dnskey in keyset.dns_keys],
            profile=NssetCheckProfile.DEFAULT,
        )
    ]


def test_start_one_missing_domain_id(db, mocker: MockFixture, time_machine: TimeMachineFixture):
    # TODO: Remove when DomainCheck.domain_id is no longer nullable
    zm_start_test_mock = mocker.patch.object(APP.zonemaster, "start_domain_test", return_value="Gazpacho!")
    rf = RegistryFactory()
    keyset = rf.keyset(1, dns_keys=[rf.dnskey(key=b"Gazpacho!"), rf.dnskey(key=b"Starbug!")])
    domain = rf.domain(1, fqdn="example.org", keyset=KeysetRef(id=keyset.keyset_id, handle=keyset.keyset_handle))
    domain_id_mock = mocker.patch.object(APP.domain_client, "get_domain_id", return_value=domain.domain_id)
    domain_mock = mocker.patch.object(APP.domain_client, "get_domain_info", return_value=domain)
    keyset_mock = mocker.patch.object(APP.keyset_client, "get_keyset_info", return_value=keyset)

    now = dt.datetime(2000, 1, 1, tzinfo=dt.UTC)
    time_machine.move_to(now)
    dt.datetime.now()  # start the time machine

    nameservers: list[dict] = [
        {"fqdn": "ns1.example.org", "ip_addresses": ["127.0.0.1", "::1"]},
        {"fqdn": "ns2.example.org"},
    ]
    mf = ModelFactory()
    with atomic(expire_on_commit=False) as sess:
        sess.add_all(
            [
                mf.nsset_check(1, nameservers=nameservers),
                dc := mf.domain_check("example.org", domain_id=None),
            ]
        )

    with atomic() as sess:
        sess.add(dc)
        async_to_sync(domain_checks.start_one)(dc)

    with atomic() as sess:
        sess.add(dc)
        sess.refresh(dc)
        assert dc.status == DomainCheckStatus.STARTED
        assert dc.test_id == "Gazpacho!"
        assert dc.started_at.timestamp() == pytest.approx(now.timestamp(), abs=1)

    assert domain_id_mock.mock_calls == [call("example.org")]
    assert domain_mock.mock_calls == [call(domain.domain_id)]
    assert keyset_mock.mock_calls == [call(keyset.keyset_id)]
    assert zm_start_test_mock.mock_calls == [
        call(
            fqdn="example.org",
            nameservers=[Nameserver(**ns) for ns in nameservers],
            ds_info=[DSInfo.from_ds_record(dnskey.ds_record("example.org")) for dnskey in keyset.dns_keys],
            profile=NssetCheckProfile.DEFAULT,
        )
    ]


def test_start_one_no_keyset(db, mocker: MockFixture, time_machine: TimeMachineFixture):
    zm_start_test_mock = mocker.patch.object(APP.zonemaster, "start_domain_test", return_value="Gazpacho!")
    rf = RegistryFactory()
    domain = rf.domain(1, fqdn="example.org")
    domain_mock = mocker.patch.object(APP.domain_client, "get_domain_info", return_value=domain)

    now = dt.datetime(2000, 1, 1, tzinfo=dt.UTC)
    time_machine.move_to(now)
    dt.datetime.now()  # start the time machine

    nameservers: list[dict] = [
        {"fqdn": "ns1.example.org", "ip_addresses": ["127.0.0.1", "::1"]},
        {"fqdn": "ns2.example.org"},
    ]
    mf = ModelFactory()
    with atomic(expire_on_commit=False) as sess:
        sess.add_all(
            [
                mf.nsset_check(1, nameservers=nameservers),
                dc := mf.domain_check("example.org", domain_id=domain.domain_id),
            ]
        )

    with atomic() as sess:
        sess.add(dc)
        async_to_sync(domain_checks.start_one)(dc)

    with atomic() as sess:
        sess.add(dc)
        sess.refresh(dc)
        assert dc.status == DomainCheckStatus.STARTED
        assert dc.test_id == "Gazpacho!"
        assert dc.started_at.timestamp() == pytest.approx(now.timestamp(), abs=1)

    assert domain_mock.mock_calls == [call(domain.domain_id)]
    assert zm_start_test_mock.mock_calls == [
        call(
            fqdn="example.org",
            nameservers=[Nameserver(**ns) for ns in nameservers],
            ds_info=[],
            profile=NssetCheckProfile.DEFAULT,
        )
    ]


def test_finish(db, mocker: MockFixture):
    mock_finish_one = mocker.patch("dnscheck.process.domain_checks.finish_one")
    zonemaster_mock = mocker.patch.object(APP.zonemaster, "test_progress", side_effect=[0, 100, 42])
    checks = _prepare_start_db(started=3)

    domain_checks.finish()

    with atomic() as sess:
        sess.add(mock_finish_one.mock_calls[0].args[0])
        assert mock_finish_one.mock_calls == [call(C(DomainCheck, domain=checks[1].domain, partial=True))]
        assert zonemaster_mock.mock_calls == [call(c.test_id) for c in checks]


def test_finish_progress_error(db, mocker: MockFixture):
    mock_finish_one = mocker.patch("dnscheck.process.domain_checks.finish_one")
    zonemaster_mock = mocker.patch.object(APP.zonemaster, "test_progress", side_effect=[0, 100, ZonemasterError])
    checks = _prepare_start_db(started=3)

    domain_checks.finish()

    with atomic() as sess:
        sess.add(mock_finish_one.mock_calls[0].args[0])
        assert mock_finish_one.mock_calls == [call(C(DomainCheck, domain=checks[1].domain, partial=True))]
        assert zonemaster_mock.mock_calls == [call(c.test_id) for c in checks]

        reset_checks = (
            sess.execute(select(DomainCheck).where(DomainCheck.status == DomainCheckStatus.CREATED)).scalars().all()
        )
        assert len(reset_checks) == 1


def test_finish_get_results_error(db, mocker: MockFixture):
    mock_finish_one = mocker.patch("dnscheck.process.domain_checks.finish_one", side_effect=ZonemasterError)
    zonemaster_mock = mocker.patch.object(APP.zonemaster, "test_progress", side_effect=[0, 100, 42])
    checks = _prepare_start_db(started=3)

    domain_checks.finish()

    with atomic() as sess:
        sess.add(mock_finish_one.mock_calls[0].args[0])
        assert mock_finish_one.mock_calls == [call(C(DomainCheck, domain=checks[1].domain, partial=True))]
        assert zonemaster_mock.mock_calls == [call(c.test_id) for c in checks]

        reset_checks = (
            sess.execute(select(DomainCheck).where(DomainCheck.status == DomainCheckStatus.CREATED)).scalars().all()
        )
        assert len(reset_checks) == 1


def test_finish_one(db, mocker: MockFixture, time_machine: TimeMachineFixture):
    test_results = [
        {"level": "INFO", "message": "This is fine\n", "module": "SYSTEM", "testcase": "UNSPECIFIED"},
        {"level": "WARNING", "message": "Something went wrong\n", "module": "BASIC", "testcase": "BASIC01"},
        {"level": "CRITICAL", "message": "Something went very wrong\n", "module": "BASIC", "testcase": "BASIC02"},
        {"level": "WARNING", "message": "Something went wrong\n", "module": "BASIC", "testcase": "BASIC03"},
    ]
    zonemaster_mock = mocker.patch.object(APP.zonemaster, "get_test_results", return_value=test_results)
    now = dt.datetime(2000, 1, 1, tzinfo=dt.UTC)
    time_machine.move_to(now)
    dt.datetime.now()  # start the time machine

    mf = ModelFactory()
    with atomic(expire_on_commit=False) as sess:
        sess.add_all(
            [
                mf.nsset_check(1),
                dc := mf.domain_check("example.org", status=DomainCheckStatus.STARTED, test_id="Gazpacho!"),
            ]
        )

    with atomic() as sess:
        sess.add(dc)
        domain_checks.finish_one(dc)

    with atomic() as sess:
        sess.add(dc)
        sess.refresh(dc)
        assert dc.status == DomainCheckStatus.FINISHED
        assert dc.finished_at.timestamp() == pytest.approx(now.timestamp(), abs=1)
        assert dc.results == test_results
        assert dc.summary == {
            TestResultLevel.INFO.value: 1,
            TestResultLevel.WARNING.value: 2,
            TestResultLevel.CRITICAL.value: 1,
        }

    assert zonemaster_mock.mock_calls == [call("Gazpacho!")]
