import datetime as dt
import signal
from unittest.mock import call

import pytest
from click.testing import CliRunner
from fred_types import NssetId
from pytest_mock import MockerFixture
from regal.exceptions import NssetDoesNotExist
from sqlalchemy import func, select
from testfixtures import LogCapture

from dnscheck.cli import cli
from dnscheck.db import DomainCheck, NssetCheck, atomic
from dnscheck.settings import APP
from dnscheck.tests.utils import ModelFactory


def test_sync(mocker: MockerFixture):
    APP.settings = None  # type: ignore[assignment]
    sync_nssets = mocker.patch("dnscheck.process.nssets.sync")
    runner = CliRunner()
    result = runner.invoke(cli, ["sync"])

    assert result.exception is None
    assert result.exit_code == 0
    assert result.output == ""
    assert sync_nssets.mock_calls == [call(batch_size=1000)]


def test_sync_batch_size(mocker: MockerFixture):
    APP.settings = None  # type: ignore[assignment]
    sync_nssets = mocker.patch("dnscheck.process.nssets.sync")
    runner = CliRunner()
    result = runner.invoke(cli, ["sync", "--batch-size", "42"])

    assert result.exception is None
    assert result.exit_code == 0
    assert result.output == ""
    assert sync_nssets.mock_calls == [call(batch_size=42)]


class RunTest:
    @pytest.fixture(autouse=True)
    def _unload_settings(self):
        """Unload settings so it can be loaded on CLI startup."""
        APP.settings = None  # type: ignore[assignment]

    @pytest.fixture(autouse=True)
    def _mock_process(self, mocker: MockerFixture):
        self.domain_checks_start = mocker.patch("dnscheck.process.domain_checks.start")
        self.domain_checks_finish = mocker.patch("dnscheck.process.domain_checks.finish")
        self.nsset_checks_start = mocker.patch("dnscheck.process.nsset_checks.start")
        self.nsset_checks_finish = mocker.patch("dnscheck.process.nsset_checks.finish")

    def test_keyboard_interrupt(self, mocker: MockerFixture):
        mocker.patch("time.sleep", side_effect=[None, KeyboardInterrupt])

        with LogCapture() as logs:
            runner = CliRunner()
            result = runner.invoke(cli, ["run"])

        assert result.exception is None
        assert result.exit_code == 0
        assert result.output == ""
        assert self.domain_checks_finish.call_count == 2
        assert self.nsset_checks_start.call_count == 2
        assert self.domain_checks_start.call_count == 2
        assert self.nsset_checks_finish.call_count == 2
        logs.check(
            ("dnscheck.cli", "INFO", "Starting dnscheck daemon"),
            ("dnscheck.cli", "INFO", "Stopping dnscheck daemon"),
        )

    def test_loop_delay(self, mocker: MockerFixture):
        sleep_mock = mocker.patch("time.sleep", side_effect=KeyboardInterrupt)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", "--loop-delay", "42"])

        assert result.exception is None
        assert result.exit_code == 0
        assert result.output == ""
        sleep_mock.assert_called_once_with(42)

    def test_sigterm(self, mocker: MockerFixture):
        def _emit_sigterm():
            signal.raise_signal(signal.SIGTERM)

        self.nsset_checks_start.side_effect = _emit_sigterm

        with LogCapture() as logs:
            runner = CliRunner()
            result = runner.invoke(cli, ["run"])

        assert result.exception is None
        assert result.exit_code == 0
        assert result.output == ""
        self.domain_checks_finish.assert_called_once()
        self.nsset_checks_start.assert_called_once()  # sigterm emitted
        self.domain_checks_start.assert_not_called()
        self.nsset_checks_finish.assert_not_called()
        logs.check(
            ("dnscheck.cli", "INFO", "Starting dnscheck daemon"),
            ("dnscheck.cli", "INFO", "Captured signal SIGTERM, stopping dnscheck daemon"),
            ("dnscheck.cli", "INFO", "Dnscheck daemon has been gracefully stopped"),
        )

    def test_unexpected_error(self, mocker: MockerFixture):
        mocker.patch("time.sleep", side_effect=KeyboardInterrupt)
        self.nsset_checks_start.side_effect = ZeroDivisionError

        with LogCapture() as logs:
            runner = CliRunner()
            result = runner.invoke(cli, ["run"])

        assert result.exception is None
        assert result.exit_code == 0
        assert result.output == ""
        self.domain_checks_finish.assert_called_once()
        self.nsset_checks_start.assert_called_once()
        self.domain_checks_start.assert_called_once()
        self.nsset_checks_finish.assert_called_once()
        logs.check(
            ("dnscheck.cli", "INFO", "Starting dnscheck daemon"),
            ("dnscheck.utils", "ERROR", "Unexpected error: ZeroDivisionError()"),
            ("dnscheck.cli", "INFO", "Stopping dnscheck daemon"),
        )


def test_clean_results(db, mocker: MockerFixture):
    APP.settings = None  # type: ignore[assignment]
    mf = ModelFactory()
    with atomic() as sess:
        sess.add(mf.nsset_check(1))
        domain_checks = [
            mf.domain_check("2020.example.org", finished_at=dt.datetime(2020, 1, 1, tzinfo=dt.UTC), results={"a": 0}),
            mf.domain_check("2021.example.org", finished_at=dt.datetime(2021, 1, 1, tzinfo=dt.UTC), results={"a": 0}),
            mf.domain_check("2022.example.org", finished_at=dt.datetime(2022, 1, 1, tzinfo=dt.UTC), results={"a": 0}),
            mf.domain_check("2023.example.org", finished_at=dt.datetime(2023, 1, 1, tzinfo=dt.UTC), results={"a": 0}),
        ]
        sess.add_all(domain_checks)

    runner = CliRunner()
    result = runner.invoke(cli, ["clean-results", "--before", "2022-01-01T00:00:00Z"])

    assert result.exception is None
    assert result.exit_code == 0
    assert result.output == ""

    with atomic() as sess:
        sess.add_all(domain_checks)
        assert domain_checks[0].results is None
        assert domain_checks[1].results is None
        assert domain_checks[2].results is not None
        assert domain_checks[3].results is not None


def test_clean_checks(db, mocker: MockerFixture):
    APP.settings = None  # type: ignore[assignment]
    mf = ModelFactory()
    with atomic() as sess:
        sess.add_all(
            [
                mf.nsset_check(1, finished_at=dt.datetime(2021, 1, 1, tzinfo=dt.UTC)),
                mf.domain_check("2020.example.org", finished_at=dt.datetime(2020, 1, 1, tzinfo=dt.UTC)),
                mf.nsset_check(2, finished_at=dt.datetime(2023, 1, 1, tzinfo=dt.UTC)),
                mf.domain_check("2022.example.org", finished_at=dt.datetime(2022, 1, 1, tzinfo=dt.UTC)),
                mf.domain_check("2023.example.org", finished_at=dt.datetime(2023, 1, 1, tzinfo=dt.UTC)),
            ]
        )

    runner = CliRunner()
    result = runner.invoke(cli, ["clean-checks", "--before", "2022-01-01T00:00:00Z"])

    assert result.exception is None
    assert result.exit_code == 0
    assert result.output == "Deleted nsset checks: 1\n"

    with atomic() as sess:
        assert sess.execute(select(func.count(NssetCheck.id))).scalar_one() == 1
        assert sess.execute(select(func.count(DomainCheck.id))).scalar_one() == 2


def test_check_ok(mocker: MockerFixture):
    start_nsset_check = mocker.patch("dnscheck.process.nsset_checks.start_one")
    mocker.patch.object(APP.nsset_client, "get_nsset_id", side_effect=[NssetId("gazpacho"), NssetId("vindaloo")])
    APP.settings = None  # type: ignore[assignment]
    runner = CliRunner()
    result = runner.invoke(cli, ["check", "STARBUG1", "STARBUG2"])

    assert result.exception is None
    assert result.exit_code == 0
    assert result.output == ("INFO: Started nsset check of STARBUG1\n" "INFO: Started nsset check of STARBUG2\n")
    assert start_nsset_check.mock_calls == [call(NssetId("gazpacho")), call(NssetId("vindaloo"))]


def test_check_errors(mocker: MockerFixture):
    start_nsset_check = mocker.patch("dnscheck.process.nsset_checks.start_one")
    mocker.patch.object(APP.nsset_client, "get_nsset_id", side_effect=[NssetDoesNotExist, NssetDoesNotExist])
    APP.settings = None  # type: ignore[assignment]
    runner = CliRunner()
    result = runner.invoke(cli, ["check", "STARBUG1", "STARBUG2"])

    assert result.exception is None
    assert result.exit_code == 0
    assert result.output == ("ERROR: Nsset STARBUG1 does not exist\n" "ERROR: Nsset STARBUG2 does not exist\n")
    assert start_nsset_check.mock_calls == []
