from unittest.mock import call
from uuid import UUID

import pytest
from pytest_mock import MockFixture
from sqlalchemy import select
from sqlalchemy.pool import NullPool, Pool

from dnscheck.db import Nsset, atomic
from dnscheck.db.common import get_engine, get_session_cls

from .utils import override_settings


@pytest.mark.parametrize("db_schema", [None, "gazpacho"])
@pytest.mark.parametrize("db_poolclass", [None, NullPool])
def test_get_engine(mocker: MockFixture, db_schema: str | None, db_poolclass: Pool | None):
    get_session_cls.cache_clear()
    get_engine.cache_clear()

    create_engine = mocker.patch("dnscheck.db.common.create_engine", autospec=True)
    with override_settings(
        db_connection="postgres:///dbname", db_echo=True, db_schema=db_schema, db_poolclass=db_poolclass
    ):
        assert get_engine() is get_engine()
        assert create_engine.call_count == 1
        assert create_engine.mock_calls == [
            call("postgres:///dbname", echo=True, poolclass=db_poolclass),
            call().update_execution_options(schema_translate_map={None: db_schema}),
        ]

    get_engine.cache_clear()
    get_session_cls.cache_clear()


def test_atomic_commit(mocker: MockFixture, db):
    with atomic() as sess:
        sess.add(Nsset(nsset_id=str(UUID(int=1)), handle="NSSET"))

    with get_session_cls()() as sess:
        nsset = sess.execute(select(Nsset)).scalar_one()
        assert nsset.handle == "NSSET"


def test_atomic_rollback(mocker: MockFixture, db):
    with pytest.raises(Exception, match="Gazpacho!"):
        with atomic() as sess:
            sess.add(Nsset(nsset_id=str(UUID(int=1)), handle="NSSET"))
            raise Exception("Gazpacho!")

    with get_session_cls()() as sess:
        nssets = sess.execute(select(Nsset)).scalars()
        assert nssets.all() == []
