from unittest.mock import call, sentinel

import pytest
from pydantic import BaseModel, ValidationError
from pytest_mock import MockerFixture

import dnscheck.settings
from dnscheck.settings import APP, HttpUrlString
from dnscheck.zonemaster import ZonemasterClient

from .utils import override_settings


class AppTest:
    @pytest.fixture(autouse=True)
    def _clear_app_cache(self):
        yield
        APP.clear_cache()

    @override_settings(zonemaster_api_url="http://zonemaster.example.com/api/")
    def test_app_zonemaster(self, mocker: MockerFixture):
        spy = mocker.spy(dnscheck.settings, "ZonemasterClient")
        zonemaster = APP.zonemaster
        assert zonemaster is APP.zonemaster
        assert isinstance(zonemaster, ZonemasterClient)
        assert spy.mock_calls == [call("http://zonemaster.example.com/api/")]

    @pytest.mark.parametrize(
        "client,attr",
        [
            ("ContactClient", "contact_client"),
            ("DomainClient", "domain_client"),
            ("KeysetClient", "keyset_client"),
            ("NssetClient", "nsset_client"),
        ],
    )
    def test_app_registry_client(self, mocker: MockerFixture, client: str, attr: str):
        mocker.patch.object(APP.settings, "registry_netloc", sentinel.netloc)
        mocker.patch.object(APP.settings, "registry_ssl_cert", sentinel.ssl_cert)
        client_mock = mocker.patch(f"dnscheck.settings.{client}", return_value=sentinel.client)
        cred_mock = mocker.patch("dnscheck.settings.make_credentials", return_value=sentinel.credentials)

        assert getattr(APP, attr) == sentinel.client

        assert client_mock.mock_calls == [call(sentinel.netloc, sentinel.credentials)]
        assert cred_mock.mock_calls == [call(sentinel.ssl_cert)]

    def test_app_messenger_client_defined(self, mocker: MockerFixture):
        mocker.patch.object(APP.settings, "messenger_netloc", sentinel.netloc)
        mocker.patch.object(APP.settings, "messenger_ssl_cert", sentinel.ssl_cert)
        client_mock = mocker.patch("dnscheck.settings.EmailMessengerClient", return_value=sentinel.client)
        cred_mock = mocker.patch("dnscheck.settings.make_credentials", return_value=sentinel.credentials)

        assert APP.email_client == sentinel.client

        assert client_mock.mock_calls == [call(sentinel.netloc, sentinel.credentials)]
        assert cred_mock.mock_calls == [call(sentinel.ssl_cert)]

    def test_app_messenger_client_undefined(self, mocker: MockerFixture):
        mocker.patch.object(APP.settings, "messenger_netloc", None)
        assert APP.email_client is None

    def test_app_clear_cache(self, mocker: MockerFixture):
        """Test that all cached properties are cleared."""
        mocker.patch.object(APP.settings, "messenger_netloc", sentinel.netloc)

        zonemaster = APP.zonemaster
        contact_client = APP.contact_client
        domain_client = APP.domain_client
        keyset_client = APP.keyset_client
        nsset_client = APP.nsset_client
        email_client = APP.email_client

        APP.clear_cache()

        assert APP.zonemaster is not zonemaster
        assert APP.contact_client is not contact_client
        assert APP.domain_client is not domain_client
        assert APP.keyset_client is not keyset_client
        assert APP.nsset_client is not nsset_client
        assert APP.email_client is not email_client


class UrlModel(BaseModel):
    url: HttpUrlString


@pytest.mark.parametrize("url", ["https://example.com/api", "http://example.com/"])
def test_http_url_string_valid(url: str):
    assert UrlModel(url=url).url == url


@pytest.mark.parametrize("url", ["gazpacho!", "", "example.com"])
def test_http_url_string_invalid(url: str):
    with pytest.raises(ValidationError):
        UrlModel(url=url)
