import base64
import re
from functools import partial
from typing import Any, cast
from uuid import UUID

from fred_types import ContactHistoryId, ContactId, DomainId, KeysetId, NssetHistoryId, NssetId
from regal import (
    Contact as ContactSchema,
    DnsKey as DnsKeySchema,
    Domain as DomainSchema,
    Keyset as KeysetSchema,
    Nsset as NssetSchema,
)
from setapp.utils import override_settings as _override_settings

from dnscheck.constants import DomainCheckStatus
from dnscheck.db import DomainCheck, Nsset, NssetCheck
from dnscheck.settings import APP

override_settings = partial(_override_settings, APP)


class ModelFactory:
    def __init__(self):
        self._last: dict[type, Any] = {
            Nsset: None,
            NssetCheck: None,
            DomainCheck: None,
        }

    def nsset(self, id: int, **kwargs):
        defaults = {
            "nsset_id": UUID(int=id).hex,
            "handle": f"HANDLE-{id}",
        }
        defaults.update(kwargs)
        self._last[Nsset] = Nsset(**defaults)
        return self._last[Nsset]

    def nsset_check(self, id: int, **kwargs):
        defaults = {
            "nsset_id": UUID(int=id).hex,
            "nsset_history_id": UUID(int=1024 + id).hex,
            "handle": f"HANDLE-{id}",
            "nameservers": [],
        }
        defaults.update(kwargs)
        self._last[NssetCheck] = NssetCheck(**defaults)
        return self._last[NssetCheck]

    def domain_check(self, fqdn: str, **kwargs):
        defaults = {
            "nsset_check": self._last[NssetCheck],
            "domain_id": f"ID-{fqdn}",
            "domain": fqdn,
            "test_id": None if kwargs.get("status") == DomainCheckStatus.CREATED else f"test:{fqdn}",
        }
        defaults.update(kwargs)
        self._last[DomainCheck] = DomainCheck(**defaults)
        return self._last[DomainCheck]


class RegistryFactory:
    def __init__(self):
        self._last: dict[type, Any] = {
            ContactSchema: None,
            DomainSchema: None,
            KeysetSchema: None,
            NssetSchema: None,
            DnsKeySchema: None,
        }

    def nsset(self, id: int, **kwargs: Any) -> NssetSchema:
        defaults = {
            "nsset_id": NssetId(UUID(int=id).hex),
            "nsset_handle": f"HANDLE-{id}",
            "nsset_history_id": NssetHistoryId(UUID(int=1024 + id).hex),
            "sponsoring_registrar": "JMC",
            "dns_hosts": [],
            "technical_check_level": 2,
            "technical_contacts": [],
        }
        defaults.update(kwargs)
        self._last[NssetSchema] = NssetSchema(**defaults)  # type: ignore[arg-type]
        return cast(NssetSchema, self._last[NssetSchema])

    def contact(self, id: int, **kwargs: Any) -> ContactSchema:
        defaults = {
            "contact_id": ContactId(UUID(int=id).hex),
            "contact_handle": f"CONTACT-{id}",
            "contact_history_id": ContactHistoryId(UUID(int=1024 + id).hex),
            "name": "Arnold J. Rimmer",
            "organization": "Jupiter Mining Corporation",
        }
        defaults.update(kwargs)
        self._last[ContactSchema] = ContactSchema(**defaults)  # type: ignore[arg-type]
        return cast(ContactSchema, self._last[ContactSchema])

    def domain(self, id: int, **kwargs: Any) -> DomainSchema:
        defaults = {
            "domain_id": DomainId(UUID(int=id).hex),
            "fqdn": f"{id}.example.org",
        }
        defaults.update(**kwargs)
        self._last[DomainSchema] = DomainSchema(**defaults)  # type: ignore[arg-type]
        return cast(DomainSchema, self._last[DomainSchema])

    def keyset(self, id: int, **kwargs: Any) -> KeysetSchema:
        defaults = {
            "keyset_id": KeysetId(UUID(int=id).hex),
            "keyset_handle": f"KEYSET-{id}",
        }
        defaults.update(**kwargs)
        self._last[KeysetSchema] = KeysetSchema(**defaults)  # type: ignore[arg-type]
        return cast(KeysetSchema, self._last[KeysetSchema])

    def dnskey(self, **kwargs: Any) -> DnsKeySchema:
        key = kwargs.pop("key", b"Gazpacho!")
        defaults = {
            "flags": 256,
            "protocol": 3,
            "alg": 13,
            "key": base64.b64encode(key).decode(),
        }
        defaults.update(**kwargs)
        self._last[DnsKeySchema] = DnsKeySchema(**defaults)  # type: ignore[arg-type]
        return cast(DnsKeySchema, self._last[DnsKeySchema])


def slugify(value: str) -> str:
    return re.sub(r"[^A-Za-z0-9_]", "_", value)
