"""DNScheck Zonemaster client."""

from typing import Any, cast

import httpx

import dnscheck

from .schema import DSInfo, Nameserver


class ZonemasterError(Exception):
    """Zonemaster error."""


class ZonemasterClient:
    """Zonemaster API client."""

    def __init__(self, api_url: str) -> None:
        self.api_url = api_url
        self.request_id = 0
        self.client = httpx.Client()

    def __del__(self) -> None:
        self.client.close()

    def _call(self, method: str, /, *, params: dict) -> Any:
        """Make API request.

        Args:
            method: Zonemaster API method name.
            params: Request params.
        """
        self.request_id += 1
        request = {
            "jsonrpc": "2.0",
            "id": self.request_id,
            "method": method,
            "params": params,
        }

        response = self.client.post(self.api_url, json=request)
        response.raise_for_status()
        body = response.json()
        if "error" in body:
            raise ZonemasterError(body["error"].get("message", "Unknown error"))
        return body.get("result")

    def start_domain_test(
        self, fqdn: str, nameservers: list[Nameserver], ds_info: list[DSInfo], *, profile: str | None = None
    ) -> str:
        """Start domain test.

        Args:
            fqdn: Fully qualified domain name.
            nameservers: List of nameservers.
            ds_info: List of DS info.
            profile: Zonemaster profile name.

        Returns:
            Test id.
        """
        params = {
            "client_id": "fred-dnscheck",
            "domain": fqdn,
            "client_version": dnscheck.__version__,
            "ds_info": [ds.model_dump(mode="json") for ds in ds_info],
            "ipv6": True,
            "ipv4": True,
        }
        if profile:
            params["profile"] = profile

        params["nameservers"] = []
        for ns in nameservers:
            if ns.ip_addresses:
                cast(list, params["nameservers"]).extend([{"ns": ns.fqdn, "ip": ip} for ip in ns.ip_addresses])
            else:
                cast(list, params["nameservers"]).append({"ns": ns.fqdn})

        return cast(str, self._call("start_domain_test", params=params))

    def test_progress(self, test_id: str) -> int:
        """Fetch test progress.

        Args:
            test_id: Zonemaster test id.

        Returns:
            Percentage between 0 and 100.
        """
        return cast(int, self._call("test_progress", params={"test_id": test_id}))

    def get_test_results(self, test_id: str) -> list[dict]:
        """Get domain test results.

        Args:
            test_id: Zonemaster test id.

        Returns:
            List of test results.
        """
        result = self._call("get_test_results", params={"id": test_id, "language": "en"})
        return cast(list[dict], result["results"])
